/* Terraform - (C) 1997-1999 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include "FileIO.h"
#include "HeightFieldExport.h"
#include "HeightFieldIO.h"
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h"
#include "MathTrig.h"
#include "TFOptions.h"
#include "tf_flexarrCOORD2.h"
#include "strrep.h"


#define NO_FLOW		-987654321		// marker for 'undefined'
#ifndef MIN
# define MIN(a,b)	(a<b ? a : b)
#endif
#ifndef MAX
# define MAX(a,b)	(a>b ? a : b)
#endif 


/*
 *  constructor: initialize everything
 */
HeightFieldExport::HeightFieldExport (HeightField *HF) 
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ HeightFieldExport\n");

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldExport::HeightFieldExport");
	SanityCheck::bailout ((!HF->d_hf), "HF->d_hf==NULL", "HeightFieldExport::HeightFieldExport");

	d_HF = HF;
	d_HFO = new HeightFieldOps (d_HF);
	SanityCheck::bailout ((!d_HFO), "!HFO==NULL", "HeightFieldExport::HeightFieldExport");
}


/*
 *  destructor: clean up 
 */
HeightFieldExport::~HeightFieldExport ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- HeightFieldExport\n");
}


/*
 *  printContourLines: iterate through all the lists and items and 
 * 	print them. 
 */
void HeightFieldExport::printContourLines (TFGListFlexArr *listMain)
{
	SanityCheck::bailout ((!listMain), "listMain==NULL", "HeightFieldExport::printContourLine");

	int		limMain=listMain->length(),
			limLine;
	COORD2		*coord2;
	TFFlexArrayCoord2 *arrLine=NULL;

	for (int i=0; i<limMain; i++)
		{
		arrLine = static_cast<TFFlexArrayCoord2*>(listMain->nth(i));
		SanityCheck::bailout ((!arrLine), "arrLine==NULL", "HeightFieldExport::printContourLine");
		limLine = arrLine->getSize ();
		printf ("******* Contour line %d of %d (length=%d) ********* \n", i, limMain, limLine); fflush (stdout);

		for (int j=0; j<limLine; j++)
			{
			coord2 = static_cast<COORD2*>(arrLine->El(j));
			printf ("(%d,%d) ; ", (int)coord2->x, (int)coord2->y);
			}

		printf ("\n\n\n");
		}
}


/*
 *  exportContourLinesToPS: iterate through all the lists and items and 
 * 	print them. 
 */
void HeightFieldExport::exportContourLinesToPS (TFGListFlexArr *listMain, char *fname) 
{
	SanityCheck::bailout ((!listMain), "listMain==NULL", "HeightFieldExport::exportContourLinesToPS");
	SanityCheck::bailout ((!fname), "fname==NULL", "HeightFieldExport::exportContourLinesToPS");
	SanityCheck::bailout ((!strlen(fname)), "strlen(fname)==0", "HeightFieldExport::exportContourLinesToPS");

	int		limMain=listMain->length(),
			limLine;
	COORD2		*coord2;
	TFFlexArrayCoord2 *arrLine=NULL;
	FileIO		*fio = new FileIO (fname);
	FILE		*fptr;

	if (!fio)
		return;

	fio->open ("w");
	fptr = fio->getFilePtr ();

	if (!fptr)
		return;
	for (int i=0; i<limMain; i++)
		{
		arrLine = static_cast<TFFlexArrayCoord2*>(listMain->nth(i));
		SanityCheck::bailout ((!arrLine), "arrLine==NULL", "HeightFieldExport::printContourLine");
		limLine = arrLine->getSize ();
		printf ("******* Contour line %d of %d (length=%d) ********* \n", i, limMain, limLine); fflush (stdout);

		for (int j=0; j<limLine; j++)
			{
			coord2 = static_cast<COORD2*>(arrLine->El(j));
			printf ("(%d,%d) ; ", (int)coord2->x, (int)coord2->y);
			}

		printf ("\n\n\n");
		}

	fio->close ();
}



/* 
 * renderPov: render a HF in POVRay. This routines forks and the parent returns
 */
int HeightFieldExport::renderPOV (char *povFile, int width, 
				  float yScale, float sealevel, 
				  bool keepFiles)
{
	SanityCheck::bailout ((!povFile), "povFile==NULL", "HeightFieldExport::renderPOV");
	SanityCheck::bailout ((!strlen(povFile)), "strlen(povFile)==0", "HeightFieldExport::renderPOV");
	SanityCheck::bailout ((width<5), "width<5", "HeightFieldExport::renderPOV");
	SanityCheck::bailout ((yScale<=0), "yScale<=0", "HeightFieldExport::renderPOV");

	FileIO	*fpov = new FileIO (povFile);	// master pov file 
	if (!fpov->exists())
		{
		SanityCheck::warning (TRUE, "povFile doesn't exist", "HeightFieldExport::renderPOV");
		delete fpov;
		return -1;
		}

	int pid = fork ();
	if (pid)
		return (0);

	// declare variables
	FileIO	*fpl = NULL,			// local pov file 
		*fhf = NULL;			// height field file
	char	*buf = static_cast<char*>(malloc(256)),
		*s, *s2, *s3, *s4, 
		*oldHFname,			// old HF name
		*povOutfile;			// povray output file
	int	height = (int)(width*0.75),
		rc;
	FILE	*fsrc, 
		*fdst;
	char	rstr1[256],
		rstr2[256],
		rstr3[256];
	bool	filesOK = TRUE;

	// original filename in comments is assumed world.tga
	fhf = new FileIO (d_HF->getName()); 				// world.tga

	// determine location of working files 
	if (keepFiles)
		sprintf (buf, "%s.pov", fhf->getBasename());	// world.pov
	else
		{
		char 	*tmpdir;

		tmpdir = getenv ("TMPDIR");
		if (!tmpdir)
			tmpdir = "/tmp";							// /tmp/world.pov
		sprintf (buf, "%s/%s.pov", tmpdir, fhf->getBasename());
		}
	fpl = new FileIO (buf);

	// assign HF file name, backup old name, create File with new name 
	sprintf (buf, "%s_hf.tga", fpl->getBasename()); 	// world_hf.tga
	oldHFname = d_HF->d_name; 
	d_HF->d_name = strdup (buf);

	// assign povray output filename 
	sprintf (buf, "%s_POVout.tga", fpl->getBasename()); // world_POVout.tga
	povOutfile = strdup (buf);

	if (!(fsrc=fpov->open("r")))
		filesOK = FALSE;
	if (!(fdst=fpl->open("w")))
		filesOK = FALSE;

	if (!filesOK)
		{
		fprintf (stderr, "Problem opening files ... aborting POV render\n");
		_exit (1);
		}

	// replacement strings
	sprintf (rstr1, "%s_hf.tga", fpl->getBasename());
	sprintf (rstr2, "%f", yScale);
	if (sealevel > 0)
		sprintf (rstr3, "water_level %f", sealevel);
	else
		strcpy (rstr3, "");

	// copy and replace the terraform placeholders
	while ((s=fgets(buf, 256, fsrc)))
		{
		s2 = strrep (s, "TF_FILE", rstr1);
		s3 = strrep (s2, "TF_YSCALE", rstr2);
		s4 = strrep (s3, "TF_WATERLEVEL", rstr3);
		fputs (s4, fdst);

		if (s4 != s3)
			free (s4);
		if (s3 != s2)
			free (s3);
		if (s2 != s)
			free (s2);
		}

	fpov->close ();
	fpl->close ();

	// write HF
	HeightFieldIO		*HFIO = new HeightFieldIO (d_HF->d_name);
	HFIO->write (d_HF);
	delete HFIO;

	// exec povray and wait for it to terminate
	//printf ("%s\n", buf);
	if (TFOptions::s_POV30)
		{
		if (GlobalTrace::isSet (GlobalTrace::TRACE_DEBUG))
			sprintf (buf, "povray +W%d +H%d +O%s +i %s ", 
				width, height, povOutfile, fpl->getName());
		else
			sprintf (buf, "povray +W%d +H%d +O%s +i %s > /dev/null 2>&1", 
				width, height, povOutfile, fpl->getName());
		}
	else
		{
		if (GlobalTrace::isSet (GlobalTrace::TRACE_DEBUG))
			sprintf (buf, "povray +D +P +W%d +H%d +O%s +i %s ", 
				width, height, povOutfile, fpl->getName());
		else
			sprintf (buf, "povray +D +P +W%d +H%d +O%s +i %s > /dev/null 2>&1", 
				width, height, povOutfile, fpl->getName());
		}
	rc = system (buf);

	// clean up 
	if (!keepFiles)
		{
		unlink (d_HF->d_name);
		unlink (fpl->getName());
		unlink (povOutfile);
		}

	delete fhf;
	delete fpl;
	delete fpov;

	free (buf);
	free (povOutfile);
	free (d_HF->d_name);
	d_HF->d_name = oldHFname;

	SanityCheck::warning (rc, "povray command returned an error", "HeightFieldExport::renderPOV");

	// have to call _exit to avoid closing file descriptors/X connection
	_exit (0);
}
